/*
Copyright (C) 2016, 2017 Siep Kroonenberg

This file is part of TLaunch.

TLaunch is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TLaunch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TLaunch.  If not, see <http://www.gnu.org/licenses/>.
*/

/////////////////////////////////////////////////////////
// editor choice

// The editors dialogue is handled here, except that some of its
// functions are invoked from tlaunch.c and tla_parse.c.

// A globally defined button ed_control has as progid the global variable
// ed_progid, which is set in initialize_editors in this file.

#include "tla_utils.h"
#include "resource.h"
#include "tlaunch.h"

static wchar_t * edsel_class = L"edsel_class";

/////////////////////////////////////////////////////////
// data structures

typedef struct {
  Progid * progid;
  // per-editor GUI items for editor selection window
  HWND hw; // radio button
  HWND hwic; // icon label
  HICON hic; // icon itself
  UINT_PTR id;
} Editor;
static Editor editors[MAX_EDITORS + 1]; // one spare for custom editor
static short neds, current_edno, cust_edno;

///////////////////////////////////////////////////////
// a couple of candidates for tla_utils, except that we need them only here

static wchar_t * name_from_cmd( wchar_t * cmd ) {
  wchar_t * prog, * p, * name;
  if( !cmd ) return NULL;
  prog = prog_from_cmd( cmd );
  if( !prog ) {
    return NULL;
  } else if( !prog[0] ) {
    FREE0( prog );
    return NULL;
  }
  p = wcsrchr( prog, L'.' );
  if( p ) *p = L'\0'; // now prog has no extension any more
  p = wcsrchr( prog, L'\\' );
  if( !p ) p = prog;
  else p++;
  name = tl_wcsncpy( p, wcslen( p ));
  name[0] = towupper( name[0] );
  FREE0( prog );
  return name;
} // name_from_cmd

static wchar_t * command_from_shell_cmd( wchar_t * shell_cmd ) {
  int l;
  wchar_t * prog, * command;
  if( !shell_cmd || !shell_cmd[0] ) return NULL;
  // assume shell_cmd was a registry value written by the launcher
  // as the result of the user selecting a custom editor.
  // Just lop off the final bit to get at the command.
  // Otherwise we would have to do a lot more parsing, or
  // store more information.
  prog = prog_from_cmd( shell_cmd );
  // first, existence check
  if( prog && prog[0] && FILE_EXISTS( prog )) {
    l = wcslen( shell_cmd );
    if( l > 5 && !wcscmp( shell_cmd + l - 5, L" \"%1\"" ))
      command = tl_wcsncpy( shell_cmd, l - 5 );
    else if( l > 8 && !wcscmp( shell_cmd + l - 8, L" \"%1\" %*" ))
      command = tl_wcsncpy( shell_cmd, l - 8 );
    else
      command = tl_wcscpy( shell_cmd );
    l = wcslen( command );
    // remove trailing spaces
    while( l > 0 && command[ l - 1 ] == L' ' ) command[ --l ] = L'\0';
    if ( !command[0] ) FREE0( command );
  } else {
    command = NULL;
  }
  FREE0( prog );
  return command;
} // command_from_shell_cmd

/////////////////////////////////////////////////////////
// exported: default editor control main window

// menu item and/or button for default editor on main window
void update_mainwin_ed_control() {
  Control * ctrl = NULL;
  int i;

  // run through all controls and check for type E
  for( i=0; i<ncontrols; i++ ) {
    ctrl = controls[i];
    if( ctrl && ctrl->type && ctrl->type == L'E' ) {
      ctrl->progid = ed_progid;
      ctrl->enabled = ed_progid->command ? TRUE : FALSE;
      EnableWindow( ctrl->hw, ctrl->enabled );

      if( i>=first_button && i<first_button + nbuttons ) {
        get_icon( ctrl );
        SendMessage( ctrl->hw, BM_SETIMAGE,
            ( WPARAM )IMAGE_ICON, ( LPARAM )ctrl->hicon );
      } // if button
    } // if control for default editor
  } // for controls
} // update_mainwin_ed_control

///////////////////////////////////////////////////////

// exported:
// build editors data structure and add custom editor, both as editor and
// as progid; called by parse_ini when it starts with menu- and button specs.
// the GUI stuff here is limited to initializing the respective structure
// members to 0 / NULL
void initialize_editors() {

  int i;
  wchar_t * shcmd, * prog, * progi, *icf, *nm;
  wchar_t * exts, * p;
  Progid * pgid;

  // collect editors from progids.
  // remember first and default editor
  first_ed = NULL;
  for( i=0; i<nprogs; i++ ) {
    pgid = progids[i];
    if( !pgid->extensions || !pgid->command ) continue;
    // tex among extensions?
    exts = pgid->extensions;
    p = wcsstr( exts, L".tex" );
    if( p && ( p == exts || *( p - 1 ) == L' ' ) &&
        (*( p + 4 ) == L'\0'  || *( p + 4 ) == L' ' )) {
      editors[neds].progid = pgid;
      if( pgid->primary && !first_ed ) first_ed = pgid;
      // create name if there is none
      if( !pgid->name )
          pgid->name = name_from_cmd( pgid->command );
      if( ++neds >= MAX_EDITORS ) break;
    }
  }
  if( neds > 0 && !first_ed ) first_ed = editors[0].progid;
  // add progid and editor for custom editor. Note that
  // the progids and editors arrays each have at least one spare slot
  cust_progid = calloc_fatal( 1, sizeof( Progid ));
  progids[nprogs++] = cust_progid;
  // in case the ini file defined CUSTOMED_PROGID:
  cust_progid->progid = get_env( L"CUSTOMED_PROGID" );
  // otherwise use the #define-d value
  if( !cust_progid->progid ) cust_progid->progid = CUSTOMED_PROGID;
  cust_progid->extensions = lx_exts;
  cust_edno = neds++;
  editors[cust_edno].progid = cust_progid;

  // a working cust_progid in the registry gets priority
  get_assoc_data( cust_progid->progid, NULL, &cust_progid->shell_cmd,
      &cust_progid->iconspec, &cust_progid->name );
  // command_from_shell_cmd also checks for existence of the program
  cust_progid->command = command_from_shell_cmd( cust_progid->shell_cmd );
  if( !cust_progid->command ) {
    FREE0( cust_progid->shell_cmd );
    FREE0( cust_progid->iconspec );
    FREE0( cust_progid->name );
  } else {
    if( !cust_progid->name )
      cust_progid->name = name_from_cmd( cust_progid->command );
    // no need to synthesize an iconspec; if absent, command will be used
  }

  // currently configured editor
  current_edno = -1;
  get_assoc_data( L".tex", NULL, &shcmd, &icf, &nm );
  if( shcmd ) {
    prog = prog_from_cmd( shcmd );
    if( FILE_EXISTS( prog )) {
      // is this a configured editor?
      for( i=0; i<neds; i++ ) {
        pgid = editors[i].progid;
        if( !pgid->command ) continue; // might only happen with cust_edno
        progi = prog_from_cmd( pgid->command );
        if( same_file( prog, progi )) {
          current_edno = i;
          FREE0( progi );
          FREE0( shcmd );
          FREE0( nm );
          FREE0( icf );
          break;
        } else {
          FREE0( progi );
        }
      } // end for
      if( current_edno < 0 ) { // prog exists, but not as configured editor
        current_edno = cust_edno;
        cust_progid->shell_cmd = shcmd;
        cust_progid->command = command_from_shell_cmd( shcmd );
        // cust_progid->progid already set
        // use specified icon, or create default
        if( icf ) cust_progid->iconspec = icf;
        if( nm ) cust_progid->name = nm;
        else cust_progid->name = name_from_cmd( shcmd );
      }
      FREE0( prog );
    } else { // prog does not exist: ignore
      FREE0( prog );
      FREE0( shcmd );
      FREE0( icf );
      FREE0( nm );
      current_edno = 0;
    }
  } else { // nothing useful configured
    FREE0( icf );
    FREE0( nm );
    current_edno = 0;
  }
  ed_progid = editors[current_edno].progid; // may be an unconfigured cust_progid

  // initialize the GUI members of the Editor structures to NULL/0
  for( i=0; i<neds; i++ ) {
    editors[i].hw = NULL;
    editors[i].hwic = NULL;
    editors[i].hic = NULL;
    editors[i].id = IDC_STATIC;
  }

  ed_initdone = TRUE;
} // initialize_editors

// select new editor via file browser.
// The editor-selection window is updated elsewhere.
static BOOL pick_custom_ed( HWND hwnd ) {
  OPENFILENAME ofn;
  BOOL chosen;
  ZeroMemory( &ofn, sizeof( OPENFILENAME ));
  ofn.lStructSize = sizeof( OPENFILENAME );
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter =
      L"Programs (*.exe, *.bat, *.cmd)\0*.exe;*.bat;*.cmd)\0\0";
  ofn.nFilterIndex = 1;
  ofn.lpstrFile = calloc_fatal( MAX_PATH, sizeof( wchar_t ));
  ofn.nMaxFile = MAX_PATH;
  ofn.lpstrTitle = L"Select custom editor";
  ofn.Flags = OFN_FILEMUSTEXIST;
  ofn.lpstrDefExt = L"exe";
  chosen = GetOpenFileName( &ofn );
  if( chosen ) {
    FREE0( cust_progid->command );
    FREE0( cust_progid->shell_cmd );
    FREE0( cust_progid->name );
    FREE0( cust_progid->iconspec );
    if( wcschr( ofn.lpstrFile, L' ' ))
      cust_progid->command = tl_wcsnconcat( 3, L"\"", ofn.lpstrFile, L"\"" );
    else
      cust_progid->command = tl_wcscpy( ofn.lpstrFile );
    cust_progid->shell_cmd = tl_wcsnconcat( 2, cust_progid->command,
        L" \"%1\"" );
    cust_progid->iconspec = tl_wcsnconcat( 2, ofn.lpstrFile, L",0" );
    cust_progid->name = name_from_cmd( ofn.lpstrFile );
    current_edno = cust_edno;
  }
  FREE0( ofn.lpstrFile );
  return chosen;
} // pick_custom_ed

/////////////////////////////////////////////////////////
// building GUI

static HWND add_editsel_child( wchar_t * cls, wchar_t * text, DWORD style,
    DWORD x, DWORD y, DWORD w, DWORD h, HWND parent, LONG id ) {
  HWND hC;
  hC = CreateWindow( cls, text,
      WS_CHILD | WS_VISIBLE | style, x, y, w, h,
      parent, (HMENU)id, GetModuleHandle( NULL ), NULL );
  if( !( hC )) tldie( parent, L"Failure to create child" );
  SendMessage( hC, WM_SETFONT, ( WPARAM )hfnt, TRUE );
  return hC;
} // add_editsel_child

static void update_ed_item( short ied ) {
  // text and icon;
  // the icon label and radio button have already been created
  wchar_t * iconfile = NULL;
  UINT iconindex = 0;
  Editor * ed = NULL;
  Progid * pgid = NULL;

  ed = editors + ied; // *( editors[ied] )
  pgid = ed->progid;
  // text
  if( pgid->name )
    SetWindowText( ed->hw, pgid->name );
  // icon: discard previous icon, if any
  if( ed->hic ) {
    SendMessage( ed->hwic, STM_SETIMAGE, ( WPARAM )IMAGE_ICON,
        ( LPARAM )NULL );
    DestroyIcon( ed->hic );
    ed->hic = NULL;
  }
  // icon: find new one
  if( pgid->iconspec ) {
    parse_iconspec( pgid->iconspec, &iconfile, &iconindex );
    if( iconfile ) {
      ed->hic = ExtractIcon( hInst, iconfile, iconindex );
      FREE0( iconfile );
    }
  }
  if( !ed->hic ) ed->hic = LoadIcon( hInst, MAKEINTRESOURCE( IDI_DEFLT ));

  // icon: apply it
  SendMessage( ed->hwic, STM_SETIMAGE, ( WPARAM )IMAGE_ICON,
      ( LPARAM )ed->hic );

} // update_ed_item

// update default editor and states of radio buttons
static void update_selected ( ) {
  int i;
  ed_progid = editors[current_edno].progid;
  for( i=0; i < neds; i++ ) {
    if( i == current_edno ) {
      SendMessage( editors[i].hw, BM_SETCHECK, 1, 0 );
      ed_progid = editors[i].progid;
    } else {
      SendMessage( editors[i].hw, BM_SETCHECK, 0, 0 );
    }
  } // end for
  if( !cust_progid || !cust_progid->command ) {
    ShowWindow( editors[cust_edno].hw, SW_HIDE );
    EnableWindow( editors[cust_edno].hw, FALSE );
    ShowWindow( editors[cust_edno].hwic, SW_HIDE );
  } else {
    ShowWindow( editors[cust_edno].hw, SW_SHOW );
    EnableWindow( editors[cust_edno].hw, TRUE );
    ShowWindow( editors[cust_edno].hwic, SW_SHOW );
  }
} // update_selected

// exported; this is the function that invokes the editor choice dialog
void ed_sel_tl( HWND hwnd ) {
  HWND hEds;
  hEds = CreateWindowEx(
      WS_EX_TOPMOST,
      edsel_class,
      L"Editor Choice",
      WS_POPUP | WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
          WS_VISIBLE,
      CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
      hMain, NULL, NULL, NULL );
  if ( hEds ) {
    ShowWindow( hEds, SW_SHOW);
    UpdateWindow( hEds);
    EnableWindow( hwnd, FALSE );
  } else {
    tlwarn( hwnd, L"Creation Editor Window Failed" );
  }
} // ed_sel_tl

static void eds_finishup ( ) {
  int i;
  EnableWindow( hMain, TRUE );
  // no need to deallocate text labels: they are either constants
  // or members of persistent data structures.
  // But we do need to deallocate icon handles
  for( i=0; i<neds; i++ ) {
    if( editors[i].hic ) {
      // make sure the icon is no longer in use
      SendMessage( editors[i].hwic, BM_SETIMAGE,
          ( WPARAM )IMAGE_ICON, ( LPARAM )NULL );
      DestroyIcon( editors[i].hic );
      editors[i].hic = NULL;
    } // if
  } // for
  if( assoc_mode ) {
    for( i=0; i<NEXTS; i++ ) set_assoc_abs( latex_exts[i], ed_progid->progid );
  }
  register_progid( cust_progid->progid, cust_progid->shell_cmd,
      cust_progid->iconspec, cust_progid->name );
  update_mainwin_ed_control();
} // eds_finishup

LRESULT CALLBACK EdsProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {

  int i, j;
  HWND hClose, hCustom;
  UINT_PTR id_next;
  Progid * pgid;
  switch( msg ) {
    case WM_CREATE: {
      // editor data structures already created
      int y, w, h;
      RECT rect = { -1, -1, -1, -1 };
      int lh = cy / 2 + ( iconsize > ( 3 * cy / 2 ) ? iconsize : 3 * cy / 2 );

      if( GetWindowRect( hMain, &rect ))
        SetWindowPos( hwnd, HWND_TOP, rect.left + 30, rect.top + 30,
            40 * cx, (( neds>1 ) ? (( neds + 3 ) * lh + cy ) : ( 15 * cy / 2 )),
            SWP_NOACTIVATE | SWP_NOZORDER );
      else
        SetWindowPos( hwnd, HWND_TOP, 0, 0,
            40 * cx, (( neds>1 ) ? (( neds + 3 ) * lh + cy ) : ( 15 * cy / 2 )),
            SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER );

      // create some radio buttons with static icon labels
      y = cy / 2;
      w = 24 * cx;
      h = lh;
      id_next = ID_FIRST_EDS_CHILD;
      if( neds > 1 ) {
        for( i=0; i<neds; i++ ) {
          editors[i].id = id_next++;
          editors[i].hw = add_editsel_child( L"button", NULL, BS_RADIOBUTTON,
              4 * cx + iconsize, y, w, h, hwnd, editors[i].id );
          // also create a static item with an editor icon
          editors[i].hwic =  add_editsel_child( L"static", NULL, SS_ICON,
              2 * cx, y, iconsize, iconsize, hwnd, IDC_STATIC );
          update_ed_item( i );
          y += lh;
        } // for
        update_selected( );
        y += cy;
      } // otherwise no radio buttons

      // custom selection button
      hCustom = add_editsel_child( L"button", L"Custom editor...",
          BS_PUSHBUTTON | WS_TABSTOP, 2 * cx, y, w, 2 * cy, hwnd,
          ID_CUSTOM );
      if( !hCustom ) tldie( hwnd, L"Failure to create button" );
      y += 5 * cy / 2;

      // close button
      hClose = add_editsel_child( L"button", L"Close",
          BS_PUSHBUTTON | WS_TABSTOP, 2 * cx, y, w, 2 * cy, hwnd,
          ID_CLOSE_EDS );
      if( !hClose ) tldie( hwnd, L"Failure to create button" );

      return 0;
    } // end WM_CREATE

    case WM_CLOSE: {
      eds_finishup( );
      DestroyWindow( hwnd );
      return 0;
    }
    case WM_COMMAND: {
      DWORD id = LOWORD( wParam );
      // editor radio button?
      if( id >= editors[0].id && id <= editors[ neds - 1 ].id ) {
        for( j=0; j<neds; j++ ) if( id == editors[j].id ) {
          pgid = editors[j].progid;
          if( pgid->command ) {
            current_edno = j;
            break;
          } // if pgid->command
        } // for j if id
        update_selected( );
        return 0;
      } else if( id == ID_CUSTOM ) {
        // file browser
        if( !pick_custom_ed( hwnd )) return 0;
        if( cust_progid->command ) {
          current_edno = cust_edno;
          ed_progid = cust_progid;
        }
        update_ed_item( cust_edno );
        update_selected( );
        return 0;
      } else if( id == ID_CLOSE_EDS ) {
        PostMessage( hwnd, WM_CLOSE, 0, 0 );
        return 0;
      }
      break;
    } // end WM_COMMAND
  } // switch msg
  return DefWindowProc( hwnd, msg, wParam, lParam );
} // EdsProc

void register_eds_wnd_class( ) {
  WNDCLASSEX wedselc;

  wedselc.cbSize = sizeof(WNDCLASSEX);
  wedselc.style = 0; // set by CreateWindow
  wedselc.lpfnWndProc = EdsProc;
  wedselc.cbClsExtra = 0;
  wedselc.cbWndExtra = 0;
  wedselc.hInstance = hInst;
  wedselc.hIcon =
      LoadIcon( GetModuleHandle( NULL ), MAKEINTRESOURCE( IDI_TL ));
  wedselc.hCursor = NULL;
  wedselc.hbrBackground = ( HBRUSH )( COLOR_BTNFACE + 1 );
  wedselc.lpszMenuName = NULL;
  wedselc.lpszClassName = edsel_class;
  wedselc.hIconSm = NULL;

  if(!RegisterClassEx( &wedselc ))
      tldie( NULL, L"Edsel Window Registration Failed!" );
}
